/*
 * class periodicSpline - implementation
 *
 * Algorithm taken from:
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2002   Uwe Rathmann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *
 * $Author: claudio $
 *
 * $Revision: 1.2 $
 *
 * $Log: periodicspline.cpp,v $
 * Revision 1.2  2017-03-06 13:05:54  claudio
 * use unsigned for comparison
 *
 * Revision 1.1.1.1  2012-09-05 08:17:22  claudio
 * Interpolator utility classes
 *
 */

#include "periodicspline.h"
#include <cmath>

namespace Interpolator{

periodicSpline::periodicSpline (doubleVector& xvalues, doubleVector&  yvalues ):Interpolator(){
	if(xvalues.size() != yvalues.size()) throw std::length_error("periodicSpline(): data vectors must be same length");
	if(xvalues.size() < 2) throw std::length_error("periodicSpline(): data vector length must be >=2");
	n_base=xvalues.size();
	x_base = new double [n_base];
	y_base= new double [n_base];
	//from qwt_spline
	a.resize(n_base-1);
	b.resize(n_base-1);
	c.resize(n_base-1);
	
	//copy the base values into the just allocated arrays
	//and check that xvalues are sorted correctly
	double lowbound=-HUGE_VAL;
	for(unsigned int i=0;i < n_base;i++){
		if(xvalues[i]>lowbound){
			lowbound=xvalues[i];
		}
		else{
			delete [] y_base;
			delete [] x_base;
			throw std::domain_error("periodicSpline(): x_values must be in ascending order");
		}
		
		x_base[i]=xvalues[i];
		y_base[i]=yvalues[i];
	}
	//record range limits
	_min=x_base[0];
	_max=x_base[n_base-1];
	
	{
	//from qwt_spline build cofficents for periodic spline
		std::vector<double> d(n_base-1);
		std::vector<double> h(n_base-1);
		std::vector<double> s(n_base);
		 //
    //  setup equation system; use coefficient
    //  vectors as temporary buffers
    //
		for ( unsigned int i = 0; i < n_base - 1; i++)
		{
			h[i] = x_base[i+1] - x_base[i];
			if (h[i] <= 0.0)
				throw std::domain_error("periodicSpline(): x_values must be in ascending order");
		}
		
		const int imax = n_base - 2;
		double htmp = h[imax];
		double dy1 = (y_base[0] - y_base[imax]) / htmp;
		for (int i = 0; i <= imax; i++)
		{
			b[i] = c[i] = h[i];
			a[i] = 2.0 * (htmp + h[i]);
			const double dy2 = (y_base[i+1] - y_base[i]) / h[i];
			d[i] = 6.0 * ( dy1 - dy2);
			dy1 = dy2;
			htmp = h[i];
		}
		
    //
    // solve it
    //
		
    // L-U Factorization
		a[0] = sqrt(a[0]);
		c[0] = h[imax] / a[0];
		double sum = 0;
		
		for( int i = 0; i < imax - 1; i++)
		{
			b[i] /= a[i];
			if (i > 0)
				c[i] = - c[i-1] * b[i-1] / a[i];
			a[i+1] = sqrt( a[i+1] - (b[i] * b[i]));
			sum += (c[i] * c[i]);
		}
		b[imax-1] = (b[imax-1] - c[imax-2] * b[imax-2]) / a[imax-1];
		a[imax] = sqrt(a[imax] - (b[imax-1] * b[imax-1]) - sum);
		
		
    // forward elimination
		s[0] = d[0] / a[0];
		sum = 0;
		for( int i = 1; i < imax; i++)
		{
			s[i] = (d[i] - b[i-1] * s[i-1]) / a[i];
			sum += c[i-1] * s[i-1];
		}
		s[imax] = (d[imax] - b[imax-1] * s[imax-1] - sum) / a[imax];
		
		
    // backward elimination
		s[imax] = - s[imax] / a[imax];
		s[imax-1] = -(s[imax-1] + b[imax-1] * s[imax]) / a[imax-1];
		for (int i= imax - 2; i >= 0; i--)
			s[i] = - (s[i] + b[i] * s[i+1] + c[i] * s[imax]) / a[i];
		
    //
    // Finally, determine the spline coefficients
    //
		s[n_base-1] = s[0];
		for ( unsigned int i=0; i < n_base-1; i++)
		{
			a[i] = ( s[i+1] - s[i] ) / ( 6.0 * h[i]);
			b[i] = 0.5 * s[i];
			c[i] = ( y_base[i+1] - y_base[i] ) / h[i] - (s[i+1] + 2.0 * s[i] ) * h[i] / 6.0; 
		}
	}		        
	
}

periodicSpline::periodicSpline(periodicSpline& src):Interpolator()
{
	if( src.n_base < 2 ) throw std::length_error("periodicSpline(): data vector length must be >=2");
	n_base=src.n_base;
	x_base = new double [n_base];
	y_base= new double [n_base];
	_min=src._min;
	_max=src._max;
	for(unsigned int i=0; i<n_base;i++){
		x_base[i]=src.x_base[i];
		y_base[i]=src.y_base[i];
	}
	a=src.a;
	b=src.b;
	c=src.c;
}

periodicSpline& periodicSpline::operator=(const periodicSpline& src)
{
	if( src.n_base < 2 ) throw std::length_error("periodicSpline(): data vector length must be >=2");
	if(n_base){
		delete [] y_base;
		delete [] x_base;
	}
	n_base=src.n_base;
	x_base = new double [n_base];
	y_base= new double [n_base];
	_min=src._min;
	_max=src._max;
	for(unsigned int i=0; i<n_base;i++){
		x_base[i]=src.x_base[i];
		y_base[i]=src.y_base[i];
	}
	a=src.a;
	b=src.b;
	c=src.c;
	return *this;
}

periodicSpline::~periodicSpline ( )
{
	if(n_base){
		delete [] y_base;
		delete [] x_base;
	}
}

/**
 * @param  xvalues returns the abscissae of the base point of the periodicspline
 * @param  yvalues returns the ordinates of the base point of the periodicspline
 */
void periodicSpline::get_base_points (doubleVector& xvalues, doubleVector& yvalues ) const
{
	xvalues.clear();
	yvalues.clear();
	for(unsigned int i=0;i<n_base;i++){
		xvalues.push_back(x_base[i]);
		yvalues.push_back(y_base[i]);
	}
	return;
}

double periodicSpline::evaluate (double v ) throw(std::range_error)
{
	
	int klo,khi,k;
	
	if (v < _min || v > _max ) throw std::range_error("input value out of range");
	
	klo=0;
	khi=n_base-1; //era n_base !!!!!
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;//k=(khi+klo)/2;
		if (x_base[k] > v) khi=k;
		else klo=k;
	}
	
	const double delta = v - x_base[klo];
	return( ( ( ( a[klo] * delta) + b[klo] )  * delta + c[klo] ) * delta + y_base[klo]);
}

doubleVector periodicSpline::evaluate (doubleVector in_val ) throw(std::range_error)
{
	//simplicistic implementation. Should make better use of the indexing
	doubleVector outval;
	for(unsigned int i=0; i <  in_val.size(); i++) outval.push_back(evaluate(in_val[i]));
	return outval;
}

} //close namespace Interpolator
